package Dataflow;

import java.util.Enumeration;
import java.util.Vector;

public class Analysis {
    public static final boolean FORWARDS = true;
    public static final boolean BACKWARDS = false;

    public static void dumpWorklist (Queue worklist) {
	Queue tmp;
	Node n;
	System.out.print ("Worklist:");
	tmp = new Queue ();
	while (!worklist.isEmpty ()) {
	    n = (Node) worklist.remove ();
	    System.out.print (" " + n.name ());
	    tmp.add (n);
	}
	System.out.println ();
	while (!tmp.isEmpty ()) {
	    worklist.add ((Node) tmp.remove ());
	}
    }

    public static int worklistSolve (ControlFlowGraph graph, boolean verbose) 
	throws Exception {

	// Appel, p.401.
	Enumeration en;
	Vector downstream;
	Queue worklist;
	Node n;
	Value v;
	int nflows;

	worklist = new Queue ();
	en = graph.elements ();
	while (en.hasMoreElements ()) {
	    n = (Node) en.nextElement ();
	    if (!n.name ().equals ("n10")) {
		// n10 used to prime the analysis; don't operate on it
		worklist.add (n);
	    }
	}

	if (verbose) {
	    System.out.println ("=======================================\n" +
				"Beginning worklist algorithm ...\n");
	}
	nflows = 0;
	while (!worklist.isEmpty ()) {
	    if (verbose) {
		System.out.println (graph + "\n");
		dumpWorklist (worklist);
		System.out.println ();
	    }
	    n = (Node) worklist.remove ();
	    v = n.value (); // save old value for comparison
	    n.flow ();
	    nflows++;
	    if (!v.equals (n.value ())) {
		downstream = n.downstream;
		en = downstream.elements ();
		while (en.hasMoreElements ()) {
		    n = (Node) en.nextElement ();
		    if (!worklist.contains (n)) {
			// add downstream nodes to worklist if not
			// already present
			worklist.add (n);
		    }
		}
	    }
	}

	return nflows;
    }

    public static int iterateSolve (ControlFlowGraph graph, boolean verbose) 
	throws Exception {

        // Appel, p.225.
        boolean stable;
	Enumeration en;
	Node n;
	Value v;
	int nflows;

	stable = false;
	nflows = 0;
	if (verbose) {
	    System.out.println ("=======================================\n" +
				"Beginning iterative algorithm ...\n");
	}
        while (!stable) {
            stable = true;
	    en = graph.elements ();
	    while (en.hasMoreElements ()) {
                n = (Node) en.nextElement ();
		if (!en.hasMoreElements ()) {
		    // n10 primes the analysis; don't actually process it
		    continue;
		}
		if (verbose) {
		    System.out.println (graph + "\n");
		}
		v = n.value ();	// save a copy for comparison
                n.flow ();
		nflows++;
                stable = stable && v.equals (n.value ());
            }
        }

	return nflows;
    }

    public static void main (String args[]) throws Exception {
	int iflows, wflows;
	boolean verbose = false;
	try {
	    verbose = args[0].equals ("-verbose");
	} catch (Exception ex) {
	}

	iflows = Analysis.iterateSolve (new ControlFlowGraph (), verbose);
	wflows = Analysis.worklistSolve (new ControlFlowGraph (), verbose);
	if (verbose) {
	    System.out.println ("Iterative solution required " + iflows +
				" applications of flow function.");
	    System.out.println ("Worklist solution required " + wflows +
				" applications of flow function.");
	}
    }
}
